#include <TinyGPS++.h>
#include <NeoSWSerial.h>
#include <SPI.h>
#include <SD.h>

/* The distance to an object at which the buzzer/vibrational 
 * motor is triggered
 */
#define TRIGG_DISTANCE 25

/* The maximum latitude/longitude error used in computing
 * the user's location
 */
#define LOC_ERR 0.0002

/* Coordinate indices */
#define LAT 0
#define LONG 1

/* Commands sent via Bluetooth */
#define FRONT 0
#define LEFT 1
#define RIGHT 2
#define TURN 3
#define SILENT_MODE_ON 4
#define SILENT_MODE_OFF 5
#define RED 6
#define BLUE 7
#define GREEN 8
#define CLEAR 9

/* Color limits used in detecting RGB */
#define RGB_LIM 240
#define RGB_MAX 260

/* Baud rates */
#define BAUD_RATE 9600
#define GPS_BAUD_RATE 9600



/*----- PINS ------- */
/* Button Pins */
static const int buttonPin = 7;
static const int colorBtnPin = 8;

/* Buzzer Pin */
static const int buzzerPin = 9;

/* Vibrational Motor Pin */
static const int motorPin = 6;

/* Ultrasonic Sensor Pins */
static const int echoPin = 2, trigPin = 3;

/* GPS Pins */
static const int GPS_RXPin = 4, GPS_TXPin = 5;

/* Color Sensor Pins */
static const int S0 = A0, S1 = A1, S2 = A2, S3 = A3, colorOut = A4;



/* ----- DISTANCE FEEDBACK DATA ------- */
/* Variable which specifies if silent mode is enabled (and the
 * virbational motor should be activated) or disabled (and the
 * buzzer should be activated).
 */
int silentMode = 1;

/* Variables used to detect when the button
 * for toggling silent mode is pushed.
 */
int prevButtonRead = LOW, currButtonRead;



/* ----- NAVIGATION DATA ------- */
/* An array of latitude and longitude coordinates for the path
 * retrieved from the microSD card.
 */
float path[10][2];

/* An array of walking directions for each point on the path.
 */
int directions[10];

/* The total number of points on the path */
int nrCoords = 0;

/* The point on the path that the user should reach next */
int currPoint = 0;

/* GPS data */
TinyGPSPlus gps;
NeoSWSerial GPS_ss(GPS_RXPin, GPS_TXPin);



/* ----- COLOR DETECTION DATA ------- */
/* The last color detected by the color sensor */
int currentColor = CLEAR;

/* Variables used to detect when the button
 *  for color detection is pushed.
 */
int prevBtnColor = LOW, currBtnColor;



/* Reads data from the Color Sensor and sets currentColor
 * to red, blue, green or clear.
 */
void ColorSensor_ReadColor (void) {
  int redColor = 0;
  int greenColor = 0;
  int blueColor = 0;
  int frequency = 0;

   /* -- RED -- */
  /* Set red filtered photodiodes to be read */
  digitalWrite(S2,LOW);
  digitalWrite(S3,LOW);
  
  /* Read the output frequency and map it to RGB */
  frequency = pulseIn(colorOut, LOW);
  redColor = map(frequency, 20, 170, 255,0);
  delay(100);


   /* -- GREEN -- */
  /* Set green filtered photodiodes to be read */
  digitalWrite(S2,HIGH);
  digitalWrite(S3,HIGH);
  
  /* Read the output frequency and map it to RGB*/
  frequency = pulseIn(colorOut, LOW);
  greenColor = map(frequency, 30, 210, 255,0);
  delay(100);


  /* -- BLUE -- */
  /* Set blue filtered photodiodes to be read */
  digitalWrite(S2,LOW);
  digitalWrite(S3,HIGH);

  /* Read the output frequency and map it to RGB*/
  frequency = pulseIn(colorOut, LOW);
  blueColor = map(frequency, 18, 168, 255,0);
  delay(100);


  /* Based on the readings detect the current color */
  if (redColor > RGB_LIM && redColor <= RGB_MAX){
      currentColor = RED;
  } else {
      if (blueColor > RGB_LIM && blueColor <= RGB_MAX){
           currentColor = BLUE;
        }
      else  {
        if (greenColor > RGB_LIM && greenColor <= RGB_MAX){
          currentColor = GREEN;
        } else {
          currentColor = CLEAR;
        }
      }
  }
}


/* Reads data from the Ultrasonic sensor and calculates
 * the distance to nearby objects.
 */
int UltrasonicSens_ReadDist(void) {

  /* The duration of sound wave travel */
  long duration;

  /* The distance measurement */
  int distance;

  /* Initialize the trigPin as LOW */
  digitalWrite(trigPin, LOW);
  delayMicroseconds(2);

 /* Set the trigPin HIGH for 10 microseconds */
  digitalWrite(trigPin, HIGH);
  delayMicroseconds(10);
  digitalWrite(trigPin, LOW);
  
  /* Get the sound wave travel time in microseconds */
  duration = pulseIn(echoPin, HIGH);
  
  /* Calculate the distance (speed of sound/2 - back and forth) */
  distance = duration * 0.034 / 2;
  
  return distance;
}


/* Reads geolocation data from the GPS sensor and compares
 * it to the values retrieved from the microSD. As a new point
 * from the path is reached, the walking direction is transmitted
 * to the app via Bluetooth.
 */
void GPS_ReadLocation(void) {
  
  float gps_lat = 0, gps_long = 0;
  
  while (GPS_ss.available() > 0){
      gps.encode(GPS_ss.read());

      /* When data is received fron the GPS, extract the latitude
       * and longitude coordinates.
       */
      if (gps.location.isUpdated()){
          gps_lat = gps.location.lat();
          gps_long = gps.location.lng();
          
          /* Check if the current location is within the error bounds
           * of the next point on the path. The error bounds account for
           * the lack of precision of the GPS sensor.
           */
          if ((gps_lat > path[currPoint][LAT] - LOC_ERR) && 
              (gps_lat < path[currPoint][LAT] + LOC_ERR) &&
              (gps_long > path[currPoint][LONG] - LOC_ERR) &&
              (gps_long < path[currPoint][LONG] + LOC_ERR)) {

                  /* Get the walking direction for the current point
                   * and transmit it through the Serial.
                   */
                  switch (directions[currPoint]) {
                  case FRONT:
                        Serial.print(FRONT);
                        break;
                  case LEFT:
                        Serial.print(LEFT);
                        break;
                  case RIGHT:
                        Serial.print(RIGHT);
                        break;
                  case TURN:
                        Serial.print(TURN);
                        break;
                  default:
                        break;
                }

                /* Advance to the next point on the path */
                currPoint++;
          }
        }
    }
}


/* Reads the path stored in the microSD card and stores it as a global
 * array of latitude and longitude coordinates. The walking directions
 * for each point are stored in a sepparate array.
 */
void SD_ReadAndLoadData(void){

  File myFile;
  
  /* Check if the file exists on the microSD */
  if (SD.exists("coords.txt")) {

      /* Open the file for reading */
      myFile = SD.open("coords.txt", FILE_READ);

      /* Each point is stored in the file on a separate line as:
       * 'latitude longitude direction'. 
       * The 'path' array will store coordinates for each point
       * (latitude at index 0 and longitude at index 1)
       * The 'directions' array will store the walking direction
       * for each point in 'path'.
       */
      int i=0, j=0;
      if(myFile) {
        while (myFile.available()) {
                if(j < 2) {
                  path[i][j] = myFile.parseFloat();
                  j++;
                } else if (j== 2) {
                    directions[i] = myFile.parseFloat();
                    j++;
                } else {
                  i++;
                  j = 0;
                }
         }
         /* Store the total number of points */
         nrCoords = i;

        /* Close the file */
        myFile.close();
      } 
 }
}


void setup(){

 /* Set OUTPUT pins */
 pinMode(S0, OUTPUT);
 pinMode(S1, OUTPUT);
 pinMode(S2, OUTPUT);
 pinMode(S3, OUTPUT);
 pinMode(trigPin, OUTPUT);
 pinMode(motorPin, OUTPUT);
 pinMode(buzzerPin, OUTPUT);

 /* Set INPUT pins */
 pinMode(echoPin, INPUT);
 pinMode(buttonPin, INPUT);
 pinMode(colorBtnPin, INPUT);
 pinMode(colorOut, INPUT);

 /* Set the frequency scaling of the color sensor to 20% */
 digitalWrite(S0,HIGH);
 digitalWrite(S1,LOW);

 /* Initialize the Serial */
 Serial.begin(BAUD_RATE);
 GPS_ss.begin(GPS_BAUD_RATE); 

 /* Initialize the SD module */
 if (!SD.begin(10)) {
    return;
 }
 /* Load data from the microSD card */
 SD_ReadAndLoadData();
}



void loop(){

  /* Distance reading from the ultrasonic sensor */
  int distance = 0;


 /* Reads the state of the 'silent enable' button 
  * and compares it to its previous state.
  */
 currButtonRead = digitalRead(buttonPin);
 if (currButtonRead != prevButtonRead) {

     /* Detect a button push and toggle silent Mode.
      * The change is signaled to the app via Bluetooth.
      */
     if (currButtonRead == HIGH) {
        silentMode = !silentMode;
        if (silentMode == 1) {
          Serial.print(SILENT_MODE_ON);
        }
        else {
          Serial.print(SILENT_MODE_OFF); 
        } 
     }
     delay(100);
 }
 prevButtonRead = currButtonRead;


/* Reads the state of the 'color detect' button 
 * and compares it to its previous state.
 */
 currBtnColor = digitalRead(colorBtnPin);
 if (currBtnColor != prevBtnColor) {

    /* Detect a button push and read the current color.
     * The result is sent to the app via Bluetooth.
     */
     if (currBtnColor == HIGH) { 
          ColorSensor_ReadColor();
          Serial.print(currentColor);
     } 
     delay(100);
 }
 prevBtnColor = currBtnColor;

 
  /* Read distance from the ultrasonic sensor */
  distance = UltrasonicSens_ReadDist();

  /* If the distance is smaller than the trigger
   * distance and silent mode is enabled, the 
   * vibrational motor is turned on. If silent mode
   * is disabled, the buzzer is turned on. Otherwise,
   * both are turned off.
   */
  if (distance < TRIGG_DISTANCE) {
    
    if (silentMode == 1) {
      noTone(buzzerPin);
      digitalWrite(motorPin, HIGH);
      
    } else {
      digitalWrite(motorPin, LOW);
      tone(buzzerPin, 2500);
      delay(500);
    }
  } else {
      digitalWrite(motorPin, LOW);
      noTone(buzzerPin);
   }

  /* Read GPS coordinates and check the path */
  GPS_ReadLocation();
}
